@@ -1,4 +1,4 @@
-/* $Id: memobj-r0drv-linux.c 156711 2023-04-06 09:23:01Z vgalitsy $ */
+/* $Id: memobj-r0drv-linux.c 162508 2024-03-28 11:08:47Z vgalitsy $ */
 /** @file
  * IPRT - Ring-0 Memory Objects, Linux.
  */
@@ -144,7 +144,8 @@ typedef struct RTR0MEMOBJLNX
     /** The pages in the apPages array. */
     size_t              cPages;
     /** Array of struct page pointers. (variable size) */
-    struct page        *apPages[1];
+    RT_FLEXIBLE_ARRAY_EXTENSION
+    struct page        *apPages[RT_FLEXIBLE_ARRAY];
 } RTR0MEMOBJLNX;
 /** Pointer to the linux memory object. */
 typedef RTR0MEMOBJLNX *PRTR0MEMOBJLNX;
@@ -215,9 +216,21 @@ static pgprot_t rtR0MemObjLinuxConvertProt(unsigned fProt, bool fKernel)
 #if defined(RT_ARCH_X86) || defined(RT_ARCH_AMD64)
             if (fKernel)
             {
+# if RTLNX_VER_MIN(6,6,0)
+                /* In kernel 6.6 mk_pte() macro was fortified with additional
+                 * check which does not allow to use our custom mask anymore
+                 * (see kernel commit ae1f05a617dcbc0a732fbeba0893786cd009536c).
+                 * For this particular mapping case, an existing mask PAGE_KERNEL_ROX
+                 * can be used instead. PAGE_KERNEL_ROX was introduced in
+                 * kernel 5.8, however, lets apply it for kernels 6.6 and newer
+                 * to be on a safe side.
+                 */
+                return PAGE_KERNEL_ROX;
+# else
                 pgprot_t fPg = MY_PAGE_KERNEL_EXEC;
                 pgprot_val(fPg) &= ~_PAGE_RW;
                 return fPg;
+# endif
             }
             return PAGE_READONLY_EXEC;
 #else
@@ -1174,7 +1187,11 @@ RTDECL(struct page *) rtR0MemObjLinuxVirtToPage(void *pv)
     u.Four  = *p4d_offset(&u.Global, ulAddr);
     if (RT_UNLIKELY(p4d_none(u.Four)))
         return NULL;
+#   if RTLNX_VER_MIN(5,6,0)
+    if (p4d_leaf(u.Four))
+#   else        
     if (p4d_large(u.Four))
+#   endif    
     {
         pPage = p4d_page(u.Four);
         AssertReturn(pPage, NULL);
@@ -1190,7 +1207,11 @@ RTDECL(struct page *) rtR0MemObjLinuxVirtToPage(void *pv)
     if (RT_UNLIKELY(pud_none(u.Upper)))
         return NULL;
 # if RTLNX_VER_MIN(2,6,25)
+#   if RTLNX_VER_MIN(5,6,0)
+    if (pud_leaf(u.Upper))
+#   else
     if (pud_large(u.Upper))
+#   endif    
     {
         pPage = pud_page(u.Upper);
         AssertReturn(pPage, NULL);
@@ -1206,7 +1227,11 @@ RTDECL(struct page *) rtR0MemObjLinuxVirtToPage(void *pv)
     if (RT_UNLIKELY(pmd_none(u.Middle)))
         return NULL;
 #if RTLNX_VER_MIN(2,6,0)
+#  if RTLNX_VER_MIN(5,6,0)
+    if (pmd_leaf(u.Middle))
+#  else
     if (pmd_large(u.Middle))
+#  endif    
     {
         pPage = pmd_page(u.Middle);
         AssertReturn(pPage, NULL);
@@ -1216,7 +1241,9 @@ RTDECL(struct page *) rtR0MemObjLinuxVirtToPage(void *pv)
     }
 #endif
 
-#if RTLNX_VER_MIN(2,5,5) || defined(pte_offset_map) /* As usual, RHEL 3 had pte_offset_map earlier. */
+#if RTLNX_VER_MIN(6,5,0) || RTLNX_RHEL_RANGE(9,4, 9,99)
+    pEntry = __pte_map(&u.Middle, ulAddr);
+#elif RTLNX_VER_MIN(2,5,5) || defined(pte_offset_map) /* As usual, RHEL 3 had pte_offset_map earlier. */
     pEntry = pte_offset_map(&u.Middle, ulAddr);
 #else
     pEntry = pte_offset(&u.Middle, ulAddr);
@@ -1290,7 +1317,9 @@ DECLHIDDEN(int) rtR0MemObjNativeLockUser(PPRTR0MEMOBJINTERNAL ppMem, RTR3PTR R3P
     IPRT_LINUX_SAVE_EFL_AC();
     const int cPages = cb >> PAGE_SHIFT;
     struct task_struct *pTask = rtR0ProcessToLinuxTask(R0Process);
+# if GET_USER_PAGES_API < KERNEL_VERSION(6, 5, 0)
     struct vm_area_struct **papVMAs;
+# endif
     PRTR0MEMOBJLNX  pMemLnx;
     int             rc      = VERR_NO_MEMORY;
     int  const      fWrite  = fAccess & RTMEM_PROT_WRITE ? 1 : 0;
@@ -1314,9 +1343,11 @@ DECLHIDDEN(int) rtR0MemObjNativeLockUser(PPRTR0MEMOBJINTERNAL ppMem, RTR3PTR R3P
         return VERR_NO_MEMORY;
     }
 
+# if GET_USER_PAGES_API < KERNEL_VERSION(6, 5, 0)
     papVMAs = (struct vm_area_struct **)RTMemAlloc(sizeof(*papVMAs) * cPages);
     if (papVMAs)
     {
+# endif
         LNX_MM_DOWN_READ(pTask->mm);
 
         /*
@@ -1335,8 +1366,11 @@ DECLHIDDEN(int) rtR0MemObjNativeLockUser(PPRTR0MEMOBJINTERNAL ppMem, RTR3PTR R3P
                                 fWrite,                 /* Write to memory. */
                                 fWrite,                 /* force write access. */
 # endif
-                                &pMemLnx->apPages[0],   /* Page array. */
-                                papVMAs);               /* vmas */
+                                &pMemLnx->apPages[0]    /* Page array. */
+# if GET_USER_PAGES_API < KERNEL_VERSION(6, 5, 0)
+                                , papVMAs               /* vmas */
+# endif
+                                );
         /*
          * Actually this should not happen at the moment as call this function
          * only for our own process.
@@ -1357,8 +1391,10 @@ DECLHIDDEN(int) rtR0MemObjNativeLockUser(PPRTR0MEMOBJINTERNAL ppMem, RTR3PTR R3P
                                 fWrite,                 /* Write to memory. */
                                 fWrite,                 /* force write access. */
 # endif
-                                &pMemLnx->apPages[0],   /* Page array. */
-                                papVMAs                 /* vmas */
+                                &pMemLnx->apPages[0]    /* Page array. */
+# if GET_USER_PAGES_API < KERNEL_VERSION(6, 5, 0)
+                                , papVMAs               /* vmas */
+# endif
 # if GET_USER_PAGES_API >= KERNEL_VERSION(4, 10, 0)
                                 , NULL                  /* locked */
 # endif
@@ -1377,8 +1413,11 @@ DECLHIDDEN(int) rtR0MemObjNativeLockUser(PPRTR0MEMOBJINTERNAL ppMem, RTR3PTR R3P
                                 fWrite,                 /* Write to memory. */
                                 fWrite,                 /* force write access. */
 # endif
-                                &pMemLnx->apPages[0],   /* Page array. */
-                                papVMAs);               /* vmas */
+                                &pMemLnx->apPages[0]    /* Page array. */
+# if GET_USER_PAGES_API < KERNEL_VERSION(6, 5, 0)
+                                , papVMAs               /* vmas */
+# endif
+                                );
 #endif /* GET_USER_PAGES_API < KERNEL_VERSION(4, 6, 0) */
         if (rc == cPages)
         {
@@ -1401,16 +1440,20 @@ DECLHIDDEN(int) rtR0MemObjNativeLockUser(PPRTR0MEMOBJINTERNAL ppMem, RTR3PTR R3P
             while (rc-- > 0)
             {
                 flush_dcache_page(pMemLnx->apPages[rc]);
-#if RTLNX_VER_MIN(6,3,0)
+# if GET_USER_PAGES_API < KERNEL_VERSION(6, 5, 0)
+#  if RTLNX_VER_MIN(6,3,0)
                 vm_flags_set(papVMAs[rc], VM_DONTCOPY | VM_LOCKED);
-#else
+#  else
                 papVMAs[rc]->vm_flags |= VM_DONTCOPY | VM_LOCKED;
-#endif
+#  endif
+# endif
             }
 
             LNX_MM_UP_READ(pTask->mm);
 
+# if GET_USER_PAGES_API < KERNEL_VERSION(6, 5, 0)
             RTMemFree(papVMAs);
+# endif
 
             pMemLnx->Core.u.Lock.R0Process = R0Process;
             pMemLnx->cPages = cPages;
@@ -1437,9 +1480,12 @@ DECLHIDDEN(int) rtR0MemObjNativeLockUser(PPRTR0MEMOBJINTERNAL ppMem, RTR3PTR R3P
 
         LNX_MM_UP_READ(pTask->mm);
 
-        RTMemFree(papVMAs);
         rc = VERR_LOCK_FAILED;
+
+# if GET_USER_PAGES_API < KERNEL_VERSION(6, 5, 0)
+        RTMemFree(papVMAs);
     }
+# endif
 
     rtR0MemObjDelete(&pMemLnx->Core);
     IPRT_LINUX_RESTORE_EFL_AC();
